#include <stdio.h>
#include <stdint.h>
#include <ctype.h>
#include "error_stuff.h"
#include "util.h"

uint8_t swap_table[0x100];

// SwapByteBits from GHIII FSB Decryptor v1.0 by Invo
uint8_t SwapByteBits(uint8_t cInput)
{
   uint8_t nResult=0;

   for(int i=0; i<8; i++)
   {
      nResult = nResult << 1;
      nResult |= (cInput & 1);
      cInput = cInput >> 1;
   }

   return (nResult);
}

void init_swap_table(void)
{
    for (int i = 0; i < 0x100; i++)
    {
        swap_table[i] = SwapByteBits(i);
    }
}

static inline uint_fast32_t read_32_le_xor(const uint8_t *bytes, const long offset, const uint8_t *key, const int key_length)
{
    uint_fast32_t b1,b2,b3,b4;
    int key_offset = offset % key_length;
    b1 = bytes[offset + 0] ^ key[key_offset ++];
    if (key_offset == key_length) key_offset = 0;
    b2 = bytes[offset + 1] ^ key[key_offset ++];
    if (key_offset == key_length) key_offset = 0;
    b3 = bytes[offset + 2] ^ key[key_offset ++];
    if (key_offset == key_length) key_offset = 0;
    b4 = bytes[offset + 3] ^ key[key_offset];

    return 
        ((uint_fast32_t)b4) << 24 |
        ((uint_fast32_t)b3) << 16 |
        ((uint_fast32_t)b2) <<  8 |
        ((uint_fast32_t)b1);
}

static inline uint_fast16_t read_16_le_xor(const uint8_t *bytes, const long offset, const uint8_t *key, const int key_length)
{
    uint_fast16_t b1,b2;
    int key_offset = offset % key_length;
    b1 = bytes[offset + 0] ^ key[key_offset ++];
    if (key_offset == key_length) key_offset = 0;
    b2 = bytes[offset + 1] ^ key[key_offset];

    return 
        ((uint_fast32_t)b2) <<  8 |
        ((uint_fast32_t)b1);
}

void analyze(FILE *infile, long whole_file_size, long max_streams);

int main(int argc, char **argv)
{
    int max_streams = 0;
    if (argc != 2)
    {
        if (argc != 3 )
        {
            printf("usage: guessfsb infile [max streams]\n");
            exit(EXIT_FAILURE);
        }

        max_streams = read_long(argv[2]);
        CHECK_ERROR(max_streams < 1, "max streams must be >= 1");
        CHECK_ERROR(max_streams > UINT32_MAX,
                "max streams must be <= UINT32_MAX");
    }
    FILE *infile = fopen(argv[1],"rb");
    CHECK_ERRNO(!infile, "error opening infile");

    CHECK_ERRNO(-1 == fseek(infile, 0, SEEK_END), "fseek");
    const long file_size = ftell(infile);
    CHECK_ERRNO(-1 == file_size, "ftell");

    analyze(infile, file_size, max_streams);

    CHECK_ERRNO(EOF == fclose(infile), "fclose infile");

    printf("done!\n");
}

void analyze(FILE *infile, long whole_file_size, long max_streams_input)
{
    uint8_t *indata = NULL;
    enum {EASY_HEADER_SIZE = 16};   // We can make good guesses for these
    enum {HEADER_SIZE = 0x30};      // FSB4
    uint8_t key[EASY_HEADER_SIZE];
    uint8_t guessed_header[EASY_HEADER_SIZE];

    init_swap_table();

    indata = malloc(whole_file_size);
    CHECK_ERRNO(!indata, "malloc");

    // dump whole file
    get_bytes_seek(0, infile, indata, whole_file_size);

    // swap everything up front
    for (long i=0; i < whole_file_size; i++)
        indata[i] = swap_table[indata[i]];

    // begin guessing
    {
        // 1. We know the first 4 (assume FSB4)

        memcpy(guessed_header, "FSB4", 4);
        for (int i=0;i<4;i++) key[i] = indata[i] ^ guessed_header[i];

        // 2. We guess a stream count, starting from 1
        //    (max out at file size/0x28 which would be the whole file filled
        //     with minimal table entries)

        uint_fast32_t max_stream_count = whole_file_size / 0x28;

        if (0 != max_streams_input )
            max_stream_count = max_streams_input;
        
        printf("Testing stream counts from 1 to %"PRIdFAST32"\n",
                max_stream_count);
        fflush(stdout);

        for (uint_fast32_t stream_count = 1;
                stream_count <= max_stream_count;
                stream_count ++ )
        {
            if (stream_count % 20 == 0)
                printf("trying %" PRIuFAST32 " streams\n", stream_count);
            fflush(stdout);

            // 3. We guess a table size, from 0x28*streams to 0xffff*streams

            uint_fast32_t max_table_size = UINT32_MAX;

            if (UINT64_C(0xffff)*stream_count < max_table_size)
            {
                max_table_size = 0xffff*stream_count;
            }

            for (uint_fast32_t table_size = 0x28*stream_count;
                table_size <= max_table_size;
                table_size ++ )
            {
                // 4. From the file, header, and table size, compute the body
                // size

                uint_fast32_t body_size =
                    whole_file_size - HEADER_SIZE - table_size;

                // 5a. If the key is < 16 bytes this should do it,
                write_32_le(stream_count, &guessed_header[4]);
                write_32_le(table_size, &guessed_header[8]);
                write_32_le(body_size, &guessed_header[0xc]);

                for (int i=4;i<EASY_HEADER_SIZE;i++)
                        key[i] = indata[i] ^ guessed_header[i];

                // 5b. we look for a repeat (could be several or none)

                for (int key_length=1; key_length < EASY_HEADER_SIZE;
                        key_length++)
                {
                    int key_length_check;
                    for (key_length_check=key_length;
                        key_length_check < EASY_HEADER_SIZE;
                        key_length_check++)
                    {
                        if (key[key_length_check % key_length] !=
                            key[key_length_check])
                            break;
                    }

                    // 5c. and use that as the key size

                    if (key_length_check == EASY_HEADER_SIZE)
                    {
                        // 6a. We test each table entry, adding up the entry and
                        // file sizes, stopping if

                        uint_fast32_t entry_total = 0;
                        uint_fast32_t file_total = 0;
                        uint_fast32_t entry_offset = HEADER_SIZE;

                        int file_num;

                        for (file_num = 0; file_num < stream_count;
                                file_num ++)
                        {
                            uint_fast16_t entry_size = 
                                read_16_le_xor(indata, entry_offset,
                                    key, key_length);
#if 0
                            printf("[%d] entry_size = %"PRIxFAST16"\n",
                                    file_num,entry_size);
#endif

                            // 6b. entry size < 0x28, entry size > table size
                            if (entry_size < 0x28 || entry_size > table_size)
                                break;

                            uint_fast32_t entry_file_size =
                                read_32_le_xor(indata, entry_offset + 0x24,
                                    key, key_length);
#if 0
                            printf("[%d] entry_file_size = %"PRIxFAST16"\n",
                                    file_num,entry_file_size);
#endif

                            // 6c. file size > body size
                            if (entry_file_size > body_size)
                                break;

                            entry_total += entry_size;
                            file_total += entry_file_size;

                            // 6d. total entry sizes > table size
                            if (entry_total > table_size)
                                break;

                            // 6e. total file sizes > body size
                            if (file_total > body_size)
                                break;

                            // :TODO:
                            // 6f. check that file name is all
                            //     printable characters?


                            entry_offset += entry_size;
                        }

#if 0
                        printf("file_num = %d\n", file_num);
                        printf("table_size = %"PRIxFAST32"\n", table_size);
                        printf("body_size = %"PRIxFAST32"\n", body_size);
                        printf("entry_total = %"PRIxFAST32"\n", entry_total);
                        printf("file_total= %"PRIxFAST32"\n", file_total);
#endif


                        // 7. Consistent in the end if
                        //    table size - sum of all entries < 0x10
                        //    body size - sum of all files < 0x800

                        if ( file_num == stream_count &&
                             table_size - entry_total <= 0x10 &&
                             body_size - file_total < 0x800)
                        {
                            // :TODO:
                            // 8. If we get something consistent with these
                            //    checks, ensure that we actually got two
                            //    checks for every byte in the key, otherwise
                            //    mark key as suspect

                            for (int i=0;i<key_length;i++)
                            {
                                printf("%02" PRIx8 " ", key[i]);
                            }
                            printf(": ");

                            printf("\"");
                            for (int i=0;i<key_length;i++)
                            {
                                if (isprint(key[i]) &&
                                    key[i] != '"' &&
                                    key[i] != '\\')
                                    printf("%c", key[i]);
                                else
                                    printf("\\x%" PRIx8, key[i]);
                            }
                            printf("\"\n");
                            fflush(stdout);
                        }
                    }   // end key length possibility
                }   // end of key guess
            }   // end of table size guess
        }   // end of stream count guess
    }   // done guessing
}
